@@ -1,22 +1,30 @@
 //! Dataset implementation based on `BTreeMap` and `BTreeSet`.
-use crate::{utils::BlankIdBijection, Quad, Triple};
-use rdf_types::BlankIdBuf;
+use crate::{utils::BTreeBijection, Quad, Triple};
+use derivative::Derivative;
+use rdf_types::AsTerm;
 use std::collections::{BTreeMap, BTreeSet};
 
 /// Graph implementation based on `BTreeMap` and `BTreeSet`.
-#[derive(Debug)]
-pub struct BTreeGraph<T: Ord = rdf_types::Term> {
-	table: BTreeMap<T, BTreeMap<T, BTreeSet<T>>>,
+#[derive(Debug, Derivative)]
+#[derivative(PartialEq(bound = "S: Ord, P: Ord, O: Ord"))]
+#[derivative(Eq(bound = "S: Ord, P: Ord, O: Ord"))]
+#[derivative(Default(bound = ""))]
+pub struct BTreeGraph<S = rdf_types::Term, P = S, O = S> {
+	table: BTreeMap<S, BTreeMap<P, BTreeSet<O>>>,
 }
 
-impl<T: Ord> BTreeGraph<T> {
+impl<S, P, O> BTreeGraph<S, P, O> {
 	/// Create a new empty `BTreeGraph`.
-	pub fn new() -> BTreeGraph<T> {
+	pub fn new() -> Self {
 		Self::default()
 	}
+}
 
+impl<S: Ord, P: Ord, O: Ord> BTreeGraph<S, P, O> {
 	/// Create a new `BTreeGraph` from another graph by consuming its triples.
-	pub fn from_graph<G: crate::SizedGraph<T>>(g: G) -> BTreeGraph<T> {
+	pub fn from_graph<G: crate::SizedGraph<Subject = S, Predicate = P, Object = O>>(
+		g: G,
+	) -> BTreeGraph<S, P, O> {
 		let mut subject_map = BTreeMap::new();
 		for (subject, predicates) in g.into_subjects() {
 			let mut predicate_map = BTreeMap::new();
@@ -31,36 +39,8 @@ impl<T: Ord> BTreeGraph<T> {
 	}
 }
 
-impl<T: Ord> Default for BTreeGraph<T> {
-	fn default() -> BTreeGraph<T> {
-		BTreeGraph {
-			table: BTreeMap::new(),
-		}
-	}
-}
-
-impl<T: Ord> crate::Graph<T> for BTreeGraph<T> {
-	type Objects<'a>
-	where
-		T: 'a,
-	= Objects<'a, T>;
-	type Predicates<'a>
-	where
-		T: 'a,
-	= Predicates<'a, T>;
-	type Subjects<'a>
-	where
-		T: 'a,
-	= Subjects<'a, T>;
-	type Triples<'a>
-	where
-		T: 'a,
-	= Iter<'a, T>;
-
-	fn triples<'a>(&'a self) -> Iter<'a, T>
-	where
-		T: 'a,
-	{
+impl<S, P, O> BTreeGraph<S, P, O> {
+	fn triples(&self) -> Iter<S, P, O> {
 		Iter {
 			subjects: self.subjects(),
 			subject: None,
@@ -70,18 +50,15 @@ impl<T: Ord> crate::Graph<T> for BTreeGraph<T> {
 		}
 	}
 
-	fn subjects<'a>(&'a self) -> Subjects<'a, T>
-	where
-		T: 'a,
-	{
+	fn subjects(&self) -> Subjects<S, P, O> {
 		Subjects {
 			it: self.table.iter(),
 		}
 	}
 
-	fn predicates<'a>(&'a self, subject: &T) -> Predicates<'a, T>
+	fn predicates(&self, subject: &S) -> Predicates<P, O>
 	where
-		T: 'a,
+		S: Ord,
 	{
 		match self.table.get(subject) {
 			Some(map) => Predicates {
@@ -91,9 +68,10 @@ impl<T: Ord> crate::Graph<T> for BTreeGraph<T> {
 		}
 	}
 
-	fn objects<'a>(&'a self, subject: &T, predicate: &T) -> Objects<'a, T>
+	fn objects(&self, subject: &S, predicate: &P) -> Objects<O>
 	where
-		T: 'a,
+		S: Ord,
+		P: Ord,
 	{
 		match self.table.get(subject) {
 			Some(map) => match map.get(predicate) {
@@ -106,7 +84,12 @@ impl<T: Ord> crate::Graph<T> for BTreeGraph<T> {
 		}
 	}
 
-	fn contains(&self, Triple(subject, predicate, object): Triple<&T, &T, &T>) -> bool {
+	fn contains(&self, Triple(subject, predicate, object): Triple<&S, &P, &O>) -> bool
+	where
+		S: Ord,
+		P: Ord,
+		O: Ord,
+	{
 		match self.table.get(subject) {
 			Some(map) => match map.get(predicate) {
 				Some(map) => map.contains(object),
@@ -115,25 +98,8 @@ impl<T: Ord> crate::Graph<T> for BTreeGraph<T> {
 			None => false,
 		}
 	}
-}
 
-impl<'a, T: 'a + Ord> std::iter::IntoIterator for &'a BTreeGraph<T> {
-	type IntoIter = Iter<'a, T>;
-	type Item = Triple<&'a T, &'a T, &'a T>;
-
-	fn into_iter(self) -> Self::IntoIter {
-		use crate::Graph;
-		self.triples()
-	}
-}
-
-impl<T: Clone + Ord> crate::SizedGraph<T> for BTreeGraph<T> {
-	type IntoObjects = IntoObjects<T>;
-	type IntoPredicates = IntoPredicates<T>;
-	type IntoSubjects = IntoSubjects<T>;
-	type IntoTriples = IntoIter<T>;
-
-	fn into_triples(self) -> IntoIter<T> {
+	fn into_triples(self) -> IntoIter<S, P, O> {
 		IntoIter {
 			subjects: self.into_subjects(),
 			subject: None,
@@ -143,13 +109,16 @@ impl<T: Clone + Ord> crate::SizedGraph<T> for BTreeGraph<T> {
 		}
 	}
 
-	fn into_subjects(self) -> IntoSubjects<T> {
+	fn into_subjects(self) -> IntoSubjects<S, P, O> {
 		IntoSubjects {
 			it: self.table.into_iter(),
 		}
 	}
 
-	fn into_predicates(mut self, subject: &T) -> IntoPredicates<T> {
+	fn into_predicates(mut self, subject: &S) -> IntoPredicates<P, O>
+	where
+		S: Ord,
+	{
 		match self.table.remove(subject) {
 			Some(map) => IntoPredicates {
 				it: Some(map.into_iter()),
@@ -158,7 +127,11 @@ impl<T: Clone + Ord> crate::SizedGraph<T> for BTreeGraph<T> {
 		}
 	}
 
-	fn into_objects(mut self, subject: &T, predicate: &T) -> IntoObjects<T> {
+	fn into_objects(mut self, subject: &S, predicate: &P) -> IntoObjects<O>
+	where
+		S: Ord,
+		P: Ord,
+	{
 		match self.table.remove(subject) {
 			Some(mut map) => match map.remove(predicate) {
 				Some(map) => IntoObjects {
@@ -171,18 +144,118 @@ impl<T: Clone + Ord> crate::SizedGraph<T> for BTreeGraph<T> {
 	}
 }
 
-impl<T: Clone + Ord> std::iter::IntoIterator for BTreeGraph<T> {
-	type IntoIter = IntoIter<T>;
-	type Item = Triple<T, T, T>;
+impl<S: Ord, P: Ord, O: Ord> crate::Graph for BTreeGraph<S, P, O> {
+	type Subject = S;
+	type Predicate = P;
+	type Object = O;
+
+	type Objects<'a>
+	where
+		Self: 'a,
+		O: 'a,
+	= Objects<'a, O>;
+	type Predicates<'a>
+	where
+		Self: 'a,
+		P: 'a,
+		O: 'a,
+	= Predicates<'a, P, O>;
+	type Subjects<'a>
+	where
+		Self: 'a,
+		S: 'a,
+		P: 'a,
+		O: 'a,
+	= Subjects<'a, S, P, O>;
+	type Triples<'a>
+	where
+		Self: 'a,
+		S: 'a,
+		P: 'a,
+		O: 'a,
+	= Iter<'a, S, P, O>;
+
+	fn triples<'a>(&'a self) -> Iter<'a, S, P, O>
+	where
+		S: 'a,
+		P: 'a,
+		O: 'a,
+	{
+		self.triples()
+	}
+
+	fn subjects<'a>(&'a self) -> Subjects<'a, S, P, O>
+	where
+		S: 'a,
+		P: 'a,
+		O: 'a,
+	{
+		self.subjects()
+	}
+
+	fn predicates<'a>(&'a self, subject: &S) -> Predicates<'a, P, O>
+	where
+		P: 'a,
+		O: 'a,
+	{
+		self.predicates(subject)
+	}
+
+	fn objects<'a>(&'a self, subject: &S, predicate: &P) -> Objects<'a, O>
+	where
+		O: 'a,
+	{
+		self.objects(subject, predicate)
+	}
+
+	fn contains(&self, triple: Triple<&S, &P, &O>) -> bool {
+		self.contains(triple)
+	}
+}
+
+impl<'a, S, P, O> std::iter::IntoIterator for &'a BTreeGraph<S, P, O> {
+	type IntoIter = Iter<'a, S, P, O>;
+	type Item = Triple<&'a S, &'a P, &'a O>;
+
+	fn into_iter(self) -> Self::IntoIter {
+		self.triples()
+	}
+}
+
+impl<S: Clone + Ord, P: Clone + Ord, O: Ord> crate::SizedGraph for BTreeGraph<S, P, O> {
+	type IntoObjects = IntoObjects<O>;
+	type IntoPredicates = IntoPredicates<P, O>;
+	type IntoSubjects = IntoSubjects<S, P, O>;
+	type IntoTriples = IntoIter<S, P, O>;
+
+	fn into_triples(self) -> IntoIter<S, P, O> {
+		self.into_triples()
+	}
+
+	fn into_subjects(self) -> IntoSubjects<S, P, O> {
+		self.into_subjects()
+	}
+
+	fn into_predicates(self, subject: &S) -> IntoPredicates<P, O> {
+		self.into_predicates(subject)
+	}
+
+	fn into_objects(self, subject: &S, predicate: &P) -> IntoObjects<O> {
+		self.into_objects(subject, predicate)
+	}
+}
+
+impl<S: Clone + Ord, P: Clone + Ord, O: Ord> std::iter::IntoIterator for BTreeGraph<S, P, O> {
+	type IntoIter = IntoIter<S, P, O>;
+	type Item = Triple<S, P, O>;
 
 	fn into_iter(self) -> Self::IntoIter {
-		use crate::SizedGraph;
 		self.into_triples()
 	}
 }
 
-impl<T: Ord> crate::MutableGraph<T> for BTreeGraph<T> {
-	fn insert(&mut self, triple: Triple<T, T, T>) {
+impl<S: Ord, P: Ord, O: Ord> crate::MutableGraph for BTreeGraph<S, P, O> {
+	fn insert(&mut self, triple: Triple<S, P, O>) {
 		let (subject, predicate, object) = triple.into_parts();
 
 		match self.table.get_mut(&subject) {
@@ -206,7 +279,7 @@ impl<T: Ord> crate::MutableGraph<T> for BTreeGraph<T> {
 		}
 	}
 
-	fn absorb<G: crate::SizedGraph<T>>(&mut self, other: G) {
+	fn absorb<G: crate::SizedGraph<Subject = S, Predicate = P, Object = O>>(&mut self, other: G) {
 		let subjects = other.into_subjects();
 
 		for (subject, predicates) in subjects {
@@ -236,12 +309,12 @@ impl<T: Ord> crate::MutableGraph<T> for BTreeGraph<T> {
 	}
 }
 
-pub struct Subjects<'a, T: Ord> {
-	it: std::collections::btree_map::Iter<'a, T, BTreeMap<T, BTreeSet<T>>>,
+pub struct Subjects<'a, S, P, O> {
+	it: std::collections::btree_map::Iter<'a, S, BTreeMap<P, BTreeSet<O>>>,
 }
 
-impl<'a, T: Ord> Iterator for Subjects<'a, T> {
-	type Item = (&'a T, Predicates<'a, T>);
+impl<'a, S, P, O> Iterator for Subjects<'a, S, P, O> {
+	type Item = (&'a S, Predicates<'a, P, O>);
 
 	fn next(&mut self) -> Option<Self::Item> {
 		self.it.next().map(|(subject, map)| {
@@ -255,12 +328,12 @@ impl<'a, T: Ord> Iterator for Subjects<'a, T> {
 	}
 }
 
-pub struct IntoSubjects<T: Ord> {
-	it: std::collections::btree_map::IntoIter<T, BTreeMap<T, BTreeSet<T>>>,
+pub struct IntoSubjects<S, P, O> {
+	it: std::collections::btree_map::IntoIter<S, BTreeMap<P, BTreeSet<O>>>,
 }
 
-impl<T: Ord> Iterator for IntoSubjects<T> {
-	type Item = (T, IntoPredicates<T>);
+impl<S, P, O> Iterator for IntoSubjects<S, P, O> {
+	type Item = (S, IntoPredicates<P, O>);
 
 	fn next(&mut self) -> Option<Self::Item> {
 		self.it.next().map(|(subject, map)| {
@@ -274,12 +347,12 @@ impl<T: Ord> Iterator for IntoSubjects<T> {
 	}
 }
 
-pub struct Predicates<'a, T: Ord> {
-	it: Option<std::collections::btree_map::Iter<'a, T, BTreeSet<T>>>,
+pub struct Predicates<'a, P, O> {
+	it: Option<std::collections::btree_map::Iter<'a, P, BTreeSet<O>>>,
 }
 
-impl<'a, T: Ord> Iterator for Predicates<'a, T> {
-	type Item = (&'a T, Objects<'a, T>);
+impl<'a, P, O> Iterator for Predicates<'a, P, O> {
+	type Item = (&'a P, Objects<'a, O>);
 
 	fn next(&mut self) -> Option<Self::Item> {
 		match &mut self.it {
@@ -296,12 +369,12 @@ impl<'a, T: Ord> Iterator for Predicates<'a, T> {
 	}
 }
 
-pub struct IntoPredicates<T: Ord> {
-	it: Option<std::collections::btree_map::IntoIter<T, BTreeSet<T>>>,
+pub struct IntoPredicates<P, O> {
+	it: Option<std::collections::btree_map::IntoIter<P, BTreeSet<O>>>,
 }
 
-impl<T: Ord> Iterator for IntoPredicates<T> {
-	type Item = (T, IntoObjects<T>);
+impl<P, O> Iterator for IntoPredicates<P, O> {
+	type Item = (P, IntoObjects<O>);
 
 	fn next(&mut self) -> Option<Self::Item> {
 		match &mut self.it {
@@ -318,12 +391,12 @@ impl<T: Ord> Iterator for IntoPredicates<T> {
 	}
 }
 
-pub struct Objects<'a, T: Ord> {
-	it: Option<std::collections::btree_set::Iter<'a, T>>,
+pub struct Objects<'a, O> {
+	it: Option<std::collections::btree_set::Iter<'a, O>>,
 }
 
-impl<'a, T: Ord> Iterator for Objects<'a, T> {
-	type Item = &'a T;
+impl<'a, O> Iterator for Objects<'a, O> {
+	type Item = &'a O;
 
 	fn next(&mut self) -> Option<Self::Item> {
 		match &mut self.it {
@@ -333,12 +406,12 @@ impl<'a, T: Ord> Iterator for Objects<'a, T> {
 	}
 }
 
-pub struct IntoObjects<T: Ord> {
-	it: Option<std::collections::btree_set::IntoIter<T>>,
+pub struct IntoObjects<O> {
+	it: Option<std::collections::btree_set::IntoIter<O>>,
 }
 
-impl<T: Ord> Iterator for IntoObjects<T> {
-	type Item = T;
+impl<O> Iterator for IntoObjects<O> {
+	type Item = O;
 
 	fn next(&mut self) -> Option<Self::Item> {
 		match &mut self.it {
@@ -348,16 +421,16 @@ impl<T: Ord> Iterator for IntoObjects<T> {
 	}
 }
 
-pub struct Iter<'a, T: Ord> {
-	subjects: Subjects<'a, T>,
-	subject: Option<&'a T>,
-	predicates: Option<Predicates<'a, T>>,
-	predicate: Option<&'a T>,
-	objects: Option<Objects<'a, T>>,
+pub struct Iter<'a, S, P, O> {
+	subjects: Subjects<'a, S, P, O>,
+	subject: Option<&'a S>,
+	predicates: Option<Predicates<'a, P, O>>,
+	predicate: Option<&'a P>,
+	objects: Option<Objects<'a, O>>,
 }
 
-impl<'a, T: Ord> Iterator for Iter<'a, T> {
-	type Item = Triple<&'a T, &'a T, &'a T>;
+impl<'a, S, P, O> Iterator for Iter<'a, S, P, O> {
+	type Item = Triple<&'a S, &'a P, &'a O>;
 
 	fn next(&mut self) -> Option<Self::Item> {
 		loop {
@@ -398,16 +471,16 @@ impl<'a, T: Ord> Iterator for Iter<'a, T> {
 	}
 }
 
-pub struct IntoIter<T: Ord> {
-	subjects: IntoSubjects<T>,
-	subject: Option<T>,
-	predicates: Option<IntoPredicates<T>>,
-	predicate: Option<T>,
-	objects: Option<IntoObjects<T>>,
+pub struct IntoIter<S, P, O> {
+	subjects: IntoSubjects<S, P, O>,
+	subject: Option<S>,
+	predicates: Option<IntoPredicates<P, O>>,
+	predicate: Option<P>,
+	objects: Option<IntoObjects<O>>,
 }
 
-impl<T: Clone + Ord> Iterator for IntoIter<T> {
-	type Item = Triple<T, T, T>;
+impl<S: Clone, P: Clone, O> Iterator for IntoIter<S, P, O> {
+	type Item = Triple<S, P, O>;
 
 	fn next(&mut self) -> Option<Self::Item> {
 		loop {
@@ -448,19 +521,62 @@ impl<T: Clone + Ord> Iterator for IntoIter<T> {
 	}
 }
 
-#[derive(Debug)]
-pub struct BTreeDataset<T: Ord = rdf_types::Term> {
-	default: BTreeGraph<T>,
-	named: BTreeMap<T, BTreeGraph<T>>,
+#[derive(Debug, Derivative)]
+#[derivative(PartialEq(bound = "S: Ord, P: Ord, O: Ord, G: Ord"))]
+#[derivative(Eq(bound = "S: Ord, P: Ord, O: Ord, G: Ord"))]
+#[derivative(Default(bound = ""))]
+pub struct BTreeDataset<S = rdf_types::Term, P = S, O = S, G = S> {
+	default: BTreeGraph<S, P, O>,
+	named: BTreeMap<G, BTreeGraph<S, P, O>>,
 }
 
-impl<T: Ord> BTreeDataset<T> {
+impl<S, P, O, G> BTreeDataset<S, P, O, G> {
 	pub fn new() -> Self {
 		Self::default()
 	}
+
+	pub fn graph(&self, id: Option<&G>) -> Option<&BTreeGraph<S, P, O>>
+	where
+		G: Ord,
+	{
+		match id {
+			Some(id) => self.named.get(id),
+			None => Some(&self.default),
+		}
+	}
+
+	pub fn graphs(&self) -> Graphs<'_, S, P, O, G> {
+		Graphs {
+			default: Some(&self.default),
+			it: self.named.iter(),
+		}
+	}
+
+	pub fn quads(&self) -> Quads<'_, S, P, O, G> {
+		Quads {
+			graphs: self.graphs(),
+			graph: None,
+			triples: None,
+		}
+	}
 }
 
-impl BTreeDataset {
+impl<S: Ord, P: Ord, O: Ord, G: Ord> BTreeDataset<S, P, O, G>
+where
+	S: AsTerm,
+	<S as AsTerm>::Iri: PartialEq,
+	<S as AsTerm>::Literal: PartialEq,
+	<S as AsTerm>::BlankId: Ord,
+	P: AsTerm<BlankId = <S as AsTerm>::BlankId>,
+	<P as AsTerm>::Iri: PartialEq,
+	<P as AsTerm>::Literal: PartialEq,
+	O: AsTerm<BlankId = <S as AsTerm>::BlankId>,
+	<O as AsTerm>::Iri: PartialEq,
+	<O as AsTerm>::Literal: PartialEq,
+	G: AsTerm<BlankId = <S as AsTerm>::BlankId>,
+	<G as AsTerm>::Iri: PartialEq,
+	<G as AsTerm>::Literal: PartialEq,
+{
 	/// Checks that there is an isomorphism between this dataset and `other`.
 	///
 	/// There is an isomorphism if there exists a blank node identifier bijection
@@ -470,101 +586,73 @@ impl BTreeDataset {
 		self.find_blank_id_bijection(other).is_some()
 	}
 
-	/// Finds a blank node identifier bijection between `self` to `other`.
+	/// Finds a blank node identifier bijection between from `self` to `other`.
 	/// If such bijection exists,
 	/// there is an isomorphism between `self` and `other`.
 	pub fn find_blank_id_bijection<'a, 'b>(
 		&'a self,
 		other: &'b Self,
-	) -> Option<BlankIdBijection<'a, 'b>> {
-		use crate::Dataset;
+	) -> Option<BTreeBijection<'a, 'b, <S as AsTerm>::BlankId, <S as AsTerm>::BlankId>> {
+		use crate::utils::isomorphism::btree::FindBTreeBlankIdBijection;
 
-		fn has_no_blank(
-			Quad(s, p, o, g): &Quad<
-				&rdf_types::Term,
-				&rdf_types::Term,
-				&rdf_types::Term,
-				&rdf_types::Term,
-			>,
+		fn has_no_blank<S: AsTerm, P: AsTerm, O: AsTerm, G: AsTerm>(
+			Quad(s, p, o, g): &Quad<&S, &P, &O, &G>,
 		) -> bool {
-			!s.is_blank()
-				&& !p.is_blank() && !o.is_blank()
-				&& !g.map(rdf_types::Term::is_blank).unwrap_or(false)
+			!s.as_term().is_blank()
+				&& !p.as_term().is_blank()
+				&& !o.as_term().is_blank()
+				&& !g.map(|g| g.as_term().is_blank()).unwrap_or(false)
 		}
 
 		let a_non_blank: BTreeSet<_> = self.quads().filter(has_no_blank).collect();
 		let b_non_blank: BTreeSet<_> = other.quads().filter(has_no_blank).collect();
 
 		if a_non_blank == b_non_blank {
-			crate::utils::find_blank_id_bijection(self, other)
+			Self::find_btree_blank_id_bijection(self, other)
 		} else {
 			None
 		}
 	}
-
-	/// Substitutes the blank node identifiers in the dataset.
-	pub fn substitute_blank_ids(self, f: impl Clone + Fn(BlankIdBuf) -> BlankIdBuf) -> Self {
-		use crate::{MutableDataset, SizedDataset};
-		let mut result = Self::new();
-
-		fn substitute_term(
-			term: rdf_types::Term,
-			f: impl Fn(BlankIdBuf) -> BlankIdBuf,
-		) -> rdf_types::Term {
-			match term {
-				rdf_types::Term::Blank(id) => rdf_types::Term::Blank(f(id)),
-				other => other,
-			}
-		}
-
-		fn substitute_quad(
-			Quad(s, p, o, g): rdf_types::GrdfQuad,
-			f: impl Clone + Fn(BlankIdBuf) -> BlankIdBuf,
-		) -> rdf_types::GrdfQuad {
-			Quad(
-				substitute_term(s, f.clone()),
-				substitute_term(p, f.clone()),
-				substitute_term(o, f.clone()),
-				g.map(|g| substitute_term(g, f)),
-			)
-		}
-
-		for quad in self.into_quads() {
-			result.insert(substitute_quad(quad, f.clone()))
-		}
-
-		result
-	}
 }
 
-impl<T: Ord> crate::Dataset<T> for BTreeDataset<T> {
-	type Graph = BTreeGraph<T>;
+impl<S: Ord, P: Ord, O: Ord, G: Ord> crate::Dataset for BTreeDataset<S, P, O, G> {
+	type Subject = S;
+	type Predicate = P;
+	type Object = O;
+	type GraphLabel = G;
+
+	type Graph = BTreeGraph<S, P, O>;
 	type Graphs<'a>
 	where
 		Self: 'a,
-		T: 'a,
-	= Graphs<'a, T>;
+		S: 'a,
+		P: 'a,
+		O: 'a,
+		G: 'a,
+	= Graphs<'a, S, P, O, G>;
 	type Quads<'a>
 	where
 		Self: 'a,
-		T: 'a,
-	= Quads<'a, T>;
+		S: 'a,
+		P: 'a,
+		O: 'a,
+	= Quads<'a, S, P, O, G>;
 
-	fn graph(&self, id: Option<&T>) -> Option<&BTreeGraph<T>> {
+	fn graph(&self, id: Option<&G>) -> Option<&BTreeGraph<S, P, O>> {
 		match id {
 			Some(id) => self.named.get(id),
 			None => Some(&self.default),
 		}
 	}
 
-	fn graphs(&self) -> Graphs<T> {
+	fn graphs(&self) -> Graphs<'_, S, P, O, G> {
 		Graphs {
 			default: Some(&self.default),
 			it: self.named.iter(),
 		}
 	}
 
-	fn quads(&self) -> Quads<T> {
+	fn quads(&self) -> Quads<'_, S, P, O, G> {
 		Quads {
 			graphs: self.graphs(),
 			graph: None,
@@ -573,13 +661,13 @@ impl<T: Ord> crate::Dataset<T> for BTreeDataset<T> {
 	}
 }
 
-pub struct Graphs<'a, T: Ord> {
-	default: Option<&'a BTreeGraph<T>>,
-	it: std::collections::btree_map::Iter<'a, T, BTreeGraph<T>>,
+pub struct Graphs<'a, S, P, O, G> {
+	default: Option<&'a BTreeGraph<S, P, O>>,
+	it: std::collections::btree_map::Iter<'a, G, BTreeGraph<S, P, O>>,
 }
 
-impl<'a, T: Ord> Iterator for Graphs<'a, T> {
-	type Item = (Option<&'a T>, &'a BTreeGraph<T>);
+impl<'a, S, P, O, G> Iterator for Graphs<'a, S, P, O, G> {
+	type Item = (Option<&'a G>, &'a BTreeGraph<S, P, O>);
 
 	fn next(&mut self) -> Option<Self::Item> {
 		if let Some(default) = self.default {
@@ -591,13 +679,13 @@ impl<'a, T: Ord> Iterator for Graphs<'a, T> {
 	}
 }
 
-pub struct GraphsMut<'a, T: Ord> {
-	default: Option<&'a mut BTreeGraph<T>>,
-	it: std::collections::btree_map::IterMut<'a, T, BTreeGraph<T>>,
+pub struct GraphsMut<'a, S, P, O, G> {
+	default: Option<&'a mut BTreeGraph<S, P, O>>,
+	it: std::collections::btree_map::IterMut<'a, G, BTreeGraph<S, P, O>>,
 }
 
-impl<'a, T: Ord> Iterator for GraphsMut<'a, T> {
-	type Item = (Option<&'a T>, &'a mut BTreeGraph<T>);
+impl<'a, S, P, O, G> Iterator for GraphsMut<'a, S, P, O, G> {
+	type Item = (Option<&'a G>, &'a mut BTreeGraph<S, P, O>);
 
 	fn next(&mut self) -> Option<Self::Item> {
 		let mut default = None;
@@ -611,17 +699,16 @@ impl<'a, T: Ord> Iterator for GraphsMut<'a, T> {
 	}
 }
 
-pub struct Quads<'a, T: Ord> {
-	graphs: Graphs<'a, T>,
-	graph: Option<&'a T>,
-	triples: Option<Iter<'a, T>>,
+pub struct Quads<'a, S, P, O, G> {
+	graphs: Graphs<'a, S, P, O, G>,
+	graph: Option<&'a G>,
+	triples: Option<Iter<'a, S, P, O>>,
 }
 
-impl<'a, T: Ord> Iterator for Quads<'a, T> {
-	type Item = Quad<&'a T, &'a T, &'a T, &'a T>;
+impl<'a, S, P, O, G> Iterator for Quads<'a, S, P, O, G> {
+	type Item = Quad<&'a S, &'a P, &'a O, &'a G>;
 
-	fn next(&mut self) -> Option<Quad<&'a T, &'a T, &'a T, &'a T>> {
-		use crate::Graph;
+	fn next(&mut self) -> Option<Quad<&'a S, &'a P, &'a O, &'a G>> {
 		loop {
 			match &mut self.triples {
 				Some(triples) => match triples.next() {
@@ -642,11 +729,13 @@ impl<'a, T: Ord> Iterator for Quads<'a, T> {
 	}
 }
 
-impl<T: Clone + Ord> crate::SizedDataset<T> for BTreeDataset<T> {
-	type IntoGraphs = IntoGraphs<T>;
-	type IntoQuads = IntoQuads<T>;
+impl<S: Clone + Ord, P: Clone + Ord, O: Ord, G: Clone + Ord> crate::SizedDataset
+	for BTreeDataset<S, P, O, G>
+{
+	type IntoGraphs = IntoGraphs<S, P, O, G>;
+	type IntoQuads = IntoQuads<S, P, O, G>;
 
-	fn into_graph(mut self, id: Option<&T>) -> Option<Self::Graph> {
+	fn into_graph(mut self, id: Option<&G>) -> Option<Self::Graph> {
 		match id {
 			Some(id) => self.named.remove(id),
 			None => Some(self.default),
@@ -669,13 +758,13 @@ impl<T: Clone + Ord> crate::SizedDataset<T> for BTreeDataset<T> {
 	}
 }
 
-pub struct IntoGraphs<T: Ord> {
-	default: Option<BTreeGraph<T>>,
-	it: std::collections::btree_map::IntoIter<T, BTreeGraph<T>>,
+pub struct IntoGraphs<S, P, O, G> {
+	default: Option<BTreeGraph<S, P, O>>,
+	it: std::collections::btree_map::IntoIter<G, BTreeGraph<S, P, O>>,
 }
 
-impl<T: Ord> Iterator for IntoGraphs<T> {
-	type Item = (Option<T>, BTreeGraph<T>);
+impl<S, P, O, G> Iterator for IntoGraphs<S, P, O, G> {
+	type Item = (Option<G>, BTreeGraph<S, P, O>);
 
 	fn next(&mut self) -> Option<Self::Item> {
 		let mut default = None;
@@ -689,17 +778,16 @@ impl<T: Ord> Iterator for IntoGraphs<T> {
 	}
 }
 
-pub struct IntoQuads<T: Clone + Ord = rdf_types::Term> {
-	graphs: IntoGraphs<T>,
-	graph: Option<T>,
-	triples: Option<IntoIter<T>>,
+pub struct IntoQuads<S: Clone = rdf_types::Term, P: Clone = S, O = S, G: Clone = S> {
+	graphs: IntoGraphs<S, P, O, G>,
+	graph: Option<G>,
+	triples: Option<IntoIter<S, P, O>>,
 }
 
-impl<T: Clone + Ord> Iterator for IntoQuads<T> {
-	type Item = Quad<T, T, T, T>;
+impl<S: Clone, P: Clone, O, G: Clone> Iterator for IntoQuads<S, P, O, G> {
+	type Item = Quad<S, P, O, G>;
 
-	fn next(&mut self) -> Option<Quad<T, T, T, T>> {
-		use crate::SizedGraph;
+	fn next(&mut self) -> Option<Quad<S, P, O, G>> {
 		loop {
 			match &mut self.triples {
 				Some(triples) => match triples.next() {
@@ -722,14 +810,17 @@ impl<T: Clone + Ord> Iterator for IntoQuads<T> {
 	}
 }
 
-impl<T: Ord> crate::MutableDataset<T> for BTreeDataset<T> {
+impl<S: Ord, P: Ord, O: Ord, G: Ord> crate::MutableDataset for BTreeDataset<S, P, O, G> {
 	type GraphsMut<'a>
 	where
 		Self: 'a,
-		T: 'a,
-	= GraphsMut<'a, T>;
+		S: 'a,
+		P: 'a,
+		O: 'a,
+		G: 'a,
+	= GraphsMut<'a, S, P, O, G>;
 
-	fn graph_mut(&mut self, id: Option<&T>) -> Option<&mut Self::Graph> {
+	fn graph_mut(&mut self, id: Option<&G>) -> Option<&mut Self::Graph> {
 		match id {
 			Some(id) => self.named.get_mut(id),
 			None => Some(&mut self.default),
@@ -743,11 +834,11 @@ impl<T: Ord> crate::MutableDataset<T> for BTreeDataset<T> {
 		}
 	}
 
-	fn insert_graph(&mut self, id: T, graph: Self::Graph) -> Option<Self::Graph> {
+	fn insert_graph(&mut self, id: G, graph: Self::Graph) -> Option<Self::Graph> {
 		self.named.insert(id, graph)
 	}
 
-	fn insert(&mut self, quad: Quad<T, T, T, T>) {
+	fn insert(&mut self, quad: Quad<S, P, O, G>) {
 		use crate::MutableGraph;
 		let (subject, predicate, object, graph_name) = quad.into_parts();
 		match self.graph_mut(graph_name.as_ref()) {
@@ -760,9 +851,11 @@ impl<T: Ord> crate::MutableDataset<T> for BTreeDataset<T> {
 		}
 	}
 
-	fn absorb<D: crate::SizedDataset<T>>(&mut self, other: D)
-	where
-		D::Graph: crate::SizedGraph<T>,
+	fn absorb<D: crate::SizedDataset<Subject = S, Predicate = P, Object = O, GraphLabel = G>>(
+		&mut self,
+		other: D,
+	) where
+		D::Graph: crate::SizedGraph,
 	{
 		use crate::MutableGraph;
 		for (id, graph) in other.into_graphs() {
@@ -776,17 +869,10 @@ impl<T: Ord> crate::MutableDataset<T> for BTreeDataset<T> {
 	}
 }
 
-impl<T: Ord> Default for BTreeDataset<T> {
-	fn default() -> BTreeDataset<T> {
-		BTreeDataset {
-			default: BTreeGraph::default(),
-			named: BTreeMap::new(),
-		}
-	}
-}
-
-impl<T: Ord> std::iter::FromIterator<Quad<T, T, T, T>> for BTreeDataset<T> {
-	fn from_iter<I: IntoIterator<Item = Quad<T, T, T, T>>>(iter: I) -> Self {
+impl<S: Ord, P: Ord, O: Ord, G: Ord> std::iter::FromIterator<Quad<S, P, O, G>>
+	for BTreeDataset<S, P, O, G>
+{
+	fn from_iter<I: IntoIterator<Item = Quad<S, P, O, G>>>(iter: I) -> Self {
 		use crate::MutableDataset;
 		let mut ds = Self::new();
 
