@@ -111,78 +111,103 @@ pub use hash_dataset::{HashDataset, HashGraph};
 ///
 /// A graph is a collection of RDF triples.
 /// It also defines a set of iterator to easily explore the graph.
-pub trait Graph<T = rdf_types::Term> {
+pub trait Graph {
+	type Subject;
+	type Predicate;
+	type Object;
+
 	/// Triple iterators.
-	type Triples<'a>: Iterator<Item = Triple<&'a T, &'a T, &'a T>>
-	where
+	type Triples<'a>: Iterator<
+		Item = Triple<&'a Self::Subject, &'a Self::Predicate, &'a Self::Object>,
+	> where
 		Self: 'a,
-		T: 'a;
+		Self::Subject: 'a,
+		Self::Predicate: 'a,
+		Self::Object: 'a;
 
 	/// Graph subjects iterator.
 	///
 	/// Each subject is given with its associated predicates (and objects).
-	type Subjects<'a>: Iterator<Item = (&'a T, Self::Predicates<'a>)>
+	type Subjects<'a>: Iterator<Item = (&'a Self::Subject, Self::Predicates<'a>)>
 	where
 		Self: 'a,
-		T: 'a;
+		Self::Subject: 'a,
+		Self::Predicate: 'a,
+		Self::Object: 'a;
 
 	/// Subject predicates iterator.
 	///
 	/// Iterate through all the predicates associated to a given subject.
 	/// Each predicate is also given with the associated objects.
-	type Predicates<'a>: Iterator<Item = (&'a T, Self::Objects<'a>)>
+	type Predicates<'a>: Iterator<Item = (&'a Self::Predicate, Self::Objects<'a>)>
 	where
 		Self: 'a,
-		T: 'a;
+		Self::Predicate: 'a,
+		Self::Object: 'a;
 
 	/// Objects iterator.
 	///
 	/// Iterate through a set of objects.
-	type Objects<'a>: Iterator<Item = &'a T>
+	type Objects<'a>: Iterator<Item = &'a Self::Object>
 	where
 		Self: 'a,
-		T: 'a;
+		Self::Object: 'a;
 
 	// Iterate through all the triples defined in the graph.
 	fn triples<'a>(&'a self) -> Self::Triples<'a>
 	where
-		T: 'a;
+		Self::Subject: 'a,
+		Self::Predicate: 'a,
+		Self::Object: 'a;
 
 	/// Iterate through all the subjects of the graph.
 	fn subjects<'a>(&'a self) -> Self::Subjects<'a>
 	where
-		T: 'a;
+		Self::Subject: 'a,
+		Self::Predicate: 'a,
+		Self::Object: 'a;
 
 	/// Iterate through all the predicates associated to the given subject.
-	fn predicates<'a>(&'a self, subject: &T) -> Self::Predicates<'a>
+	fn predicates<'a>(&'a self, subject: &Self::Subject) -> Self::Predicates<'a>
 	where
-		T: 'a;
+		Self::Predicate: 'a,
+		Self::Object: 'a;
 
 	/// Iterate through all the objects associated to the given subject and
 	/// predicate.
-	fn objects<'a>(&'a self, subject: &T, predicate: &T) -> Self::Objects<'a>
+	fn objects<'a>(
+		&'a self,
+		subject: &Self::Subject,
+		predicate: &Self::Predicate,
+	) -> Self::Objects<'a>
 	where
-		T: 'a;
+		Self::Object: 'a;
 
 	/// Checks if the given triple is defined in the graph.
-	fn contains(&self, triple: Triple<&T, &T, &T>) -> bool;
+	fn contains(&self, triple: Triple<&Self::Subject, &Self::Predicate, &Self::Object>) -> bool;
 }
 
 /// Sized gRDF graph that can be converted into iterators.
 ///
 /// Defines a set of iterators the graph can be consumed into.
-pub trait SizedGraph<T = rdf_types::Term>: Graph<T> + Sized {
+pub trait SizedGraph: Graph + Sized {
 	/// Consuming triples iterator.
-	type IntoTriples: Iterator<Item = Triple<T, T, T>>;
+	type IntoTriples: Iterator<
+		Item = Triple<
+			<Self as Graph>::Subject,
+			<Self as Graph>::Predicate,
+			<Self as Graph>::Object,
+		>,
+	>;
 
 	/// Consuming subjects iterator.
-	type IntoSubjects: Iterator<Item = (T, Self::IntoPredicates)>;
+	type IntoSubjects: Iterator<Item = (<Self as Graph>::Subject, Self::IntoPredicates)>;
 
 	/// Consuming predicates iterator.
-	type IntoPredicates: Iterator<Item = (T, Self::IntoObjects)>;
+	type IntoPredicates: Iterator<Item = (<Self as Graph>::Predicate, Self::IntoObjects)>;
 
 	/// Consuming objects iterator.
-	type IntoObjects: Iterator<Item = T>;
+	type IntoObjects: Iterator<Item = <Self as Graph>::Object>;
 
 	/// Consumes the graph and returns an iterator over its triples.
 	fn into_triples(self) -> Self::IntoTriples;
@@ -192,22 +217,42 @@ pub trait SizedGraph<T = rdf_types::Term>: Graph<T> + Sized {
 
 	/// Consumes the graph and returns an iterator over the predicates of the
 	/// given subject.
-	fn into_predicates(self, subject: &T) -> Self::IntoPredicates;
+	fn into_predicates(self, subject: &<Self as Graph>::Subject) -> Self::IntoPredicates;
 
 	/// Consumes the graph and returns an iterator over the objects of the given
 	/// subject and predicate.
-	fn into_objects(self, subject: &T, predicate: &T) -> Self::IntoObjects;
+	fn into_objects(
+		self,
+		subject: &<Self as Graph>::Subject,
+		predicate: &<Self as Graph>::Predicate,
+	) -> Self::IntoObjects;
 }
 
 /// Mutable gRDF graph.
-pub trait MutableGraph<T = rdf_types::Term>: Graph<T> {
+pub trait MutableGraph: Graph {
 	/// Insert the given triple into the graph.
-	fn insert(&mut self, triple: Triple<T, T, T>);
+	fn insert(
+		&mut self,
+		triple: Triple<
+			<Self as Graph>::Subject,
+			<Self as Graph>::Predicate,
+			<Self as Graph>::Object,
+		>,
+	);
 
 	/// Absorb the given other graph.
 	///
 	/// Adds all the triples of `other` in the graph.
-	fn absorb<G: SizedGraph<T>>(&mut self, other: G);
+	fn absorb<
+		G: SizedGraph<
+			Subject = <Self as Graph>::Subject,
+			Predicate = <Self as Graph>::Predicate,
+			Object = <Self as Graph>::Object,
+		>,
+	>(
+		&mut self,
+		other: G,
+	);
 }
 
 /// gRDF dataset.
@@ -216,29 +261,39 @@ pub trait MutableGraph<T = rdf_types::Term>: Graph<T> {
 /// It is made of a default graph and a collection of named graphs.
 ///
 /// A dataset can also be seen as a collection of [`Quad`]s.
-pub trait Dataset<T = rdf_types::Term> {
+pub trait Dataset {
+	type Subject;
+	type Predicate;
+	type Object;
+	type GraphLabel;
+
 	/// Type of graphs in the dataset.
-	type Graph: Graph<T>;
+	type Graph: Graph<Subject = Self::Subject, Predicate = Self::Predicate, Object = Self::Object>;
 
 	/// Graph iterator.
 	///
 	/// Each graph is associated to its name (if any).
-	type Graphs<'a>: Iterator<Item = (Option<&'a T>, &'a Self::Graph)>
+	type Graphs<'a>: Iterator<Item = (Option<&'a Self::GraphLabel>, &'a Self::Graph)>
 	where
 		Self: 'a,
-		T: 'a;
+		Self::GraphLabel: 'a,
+		Self::Graph: 'a;
 
 	/// Quads iterator.
-	type Quads<'a>: Iterator<Item = Quad<&'a T, &'a T, &'a T, &'a T>>
-	where
+	type Quads<'a>: Iterator<
+		Item = Quad<&'a Self::Subject, &'a Self::Predicate, &'a Self::Object, &'a Self::GraphLabel>,
+	> where
 		Self: 'a,
-		T: 'a;
+		Self::Subject: 'a,
+		Self::Predicate: 'a,
+		Self::Object: 'a,
+		Self::GraphLabel: 'a;
 
 	/// Get the graph with the given name.
 	/// Input `None` to get the default graph.
 	///
 	/// Note to implementors: the default graph should always exists.
-	fn graph(&self, id: Option<&T>) -> Option<&Self::Graph>;
+	fn graph(&self, id: Option<&Self::GraphLabel>) -> Option<&Self::Graph>;
 
 	/// Get the default graph of the dataset.
 	///
@@ -256,10 +311,15 @@ pub trait Dataset<T = rdf_types::Term> {
 	fn quads(&self) -> Self::Quads<'_>;
 
 	/// Iterate through all the subjects of the given graph.
-	fn subjects<'a>(&'a self, id: Option<&T>) -> Option<<Self::Graph as Graph<T>>::Subjects<'a>>
+	fn subjects<'a>(
+		&'a self,
+		id: Option<&Self::GraphLabel>,
+	) -> Option<<Self::Graph as Graph>::Subjects<'a>>
 	where
 		Self::Graph: 'a,
-		T: 'a,
+		Self::Subject: 'a,
+		Self::Predicate: 'a,
+		Self::Object: 'a,
 	{
 		self.graph(id).map(|graph| graph.subjects())
 	}
@@ -268,27 +328,28 @@ pub trait Dataset<T = rdf_types::Term> {
 	/// graph.
 	fn predicates<'a>(
 		&'a self,
-		id: Option<&T>,
-		subject: &T,
-	) -> Option<<Self::Graph as Graph<T>>::Predicates<'a>>
+		id: Option<&Self::GraphLabel>,
+		subject: &Self::Subject,
+	) -> Option<<Self::Graph as Graph>::Predicates<'a>>
 	where
 		Self::Graph: 'a,
-		T: 'a,
+		Self::Predicate: 'a,
+		Self::Object: 'a,
 	{
 		self.graph(id).map(|graph| graph.predicates(subject))
 	}
 
-	/// Iterate through all the objects of the given subect and predicate of the
+	/// Iterate through all the objects of the given subject and predicate of the
 	/// given graph.
 	fn objects<'a>(
 		&'a self,
-		id: Option<&T>,
-		subject: &T,
-		predicate: &T,
-	) -> Option<<Self::Graph as Graph<T>>::Objects<'a>>
+		id: Option<&Self::GraphLabel>,
+		subject: &Self::Subject,
+		predicate: &Self::Predicate,
+	) -> Option<<Self::Graph as Graph>::Objects<'a>>
 	where
 		Self::Graph: 'a,
-		T: 'a,
+		Self::Object: 'a,
 	{
 		self.graph(id)
 			.map(|graph| graph.objects(subject, predicate))
@@ -296,18 +357,20 @@ pub trait Dataset<T = rdf_types::Term> {
 }
 
 /// Sized gRDF dataset that can be converted into iterators.
-pub trait SizedDataset<T = rdf_types::Term>: Dataset<T> + Sized
+pub trait SizedDataset: Dataset + Sized
 where
-	Self::Graph: SizedGraph<T>,
+	Self::Graph: SizedGraph,
 {
 	/// Consuming graphs iterator.
-	type IntoGraphs: Iterator<Item = (Option<T>, Self::Graph)>;
+	type IntoGraphs: Iterator<Item = (Option<Self::GraphLabel>, Self::Graph)>;
 
 	/// Consuming quads iterator.
-	type IntoQuads: Iterator<Item = Quad<T, T, T, T>>;
+	type IntoQuads: Iterator<
+		Item = Quad<Self::Subject, Self::Predicate, Self::Object, Self::GraphLabel>,
+	>;
 
 	/// Consumes the dataset and returns the given graph.
-	fn into_graph(self, id: Option<&T>) -> Option<Self::Graph>;
+	fn into_graph(self, id: Option<&Self::GraphLabel>) -> Option<Self::Graph>;
 
 	/// Consumes the dataset and returns the default graph.
 	fn into_default_graph(self) -> Self::Graph {
@@ -322,7 +385,10 @@ where
 
 	/// Consumes the dataset and returns an iterator over the subjects of the
 	/// given graph.
-	fn subjects(self, id: Option<&T>) -> Option<<Self::Graph as SizedGraph<T>>::IntoSubjects> {
+	fn subjects(
+		self,
+		id: Option<&Self::GraphLabel>,
+	) -> Option<<Self::Graph as SizedGraph>::IntoSubjects> {
 		self.into_graph(id).map(|graph| graph.into_subjects())
 	}
 
@@ -330,9 +396,9 @@ where
 	/// given subject of the given graph.
 	fn predicates(
 		self,
-		id: Option<&T>,
-		subject: &T,
-	) -> Option<<Self::Graph as SizedGraph<T>>::IntoPredicates> {
+		id: Option<&Self::GraphLabel>,
+		subject: &Self::Subject,
+	) -> Option<<Self::Graph as SizedGraph>::IntoPredicates> {
 		self.into_graph(id)
 			.map(|graph| graph.into_predicates(subject))
 	}
@@ -341,29 +407,30 @@ where
 	/// given subject and predicate of the given graph.
 	fn objects(
 		self,
-		id: Option<&T>,
-		subject: &T,
-		predicate: &T,
-	) -> Option<<Self::Graph as SizedGraph<T>>::IntoObjects> {
+		id: Option<&Self::GraphLabel>,
+		subject: &Self::Subject,
+		predicate: &Self::Predicate,
+	) -> Option<<Self::Graph as SizedGraph>::IntoObjects> {
 		self.into_graph(id)
 			.map(|graph| graph.into_objects(subject, predicate))
 	}
 }
 
 /// Mutable dataset.
-pub trait MutableDataset<T = rdf_types::Term>: Dataset<T> {
+pub trait MutableDataset: Dataset {
 	/// Iterator over mutable graphs.
-	type GraphsMut<'a>: Iterator<Item = (Option<&'a T>, &'a mut Self::Graph)>
+	type GraphsMut<'a>: Iterator<Item = (Option<&'a Self::GraphLabel>, &'a mut Self::Graph)>
 	where
 		Self: 'a,
-		T: 'a;
+		Self::GraphLabel: 'a,
+		Self::Graph: 'a;
 
 	/// Get the given graph mutably.
 	///
 	/// Use the input `None` to get the default graph.
 	///
 	/// Note to implementors: the default graph should always exists.
-	fn graph_mut(&mut self, id: Option<&T>) -> Option<&mut Self::Graph>;
+	fn graph_mut(&mut self, id: Option<&Self::GraphLabel>) -> Option<&mut Self::Graph>;
 
 	/// Get the default graph mutably.
 	///
@@ -379,15 +446,27 @@ pub trait MutableDataset<T = rdf_types::Term>: Dataset<T> {
 	///
 	/// If a graph with the given name already exists,
 	/// it is replaced and the previous graph definition is returned.
-	fn insert_graph(&mut self, id: T, graph: Self::Graph) -> Option<Self::Graph>;
+	fn insert_graph(&mut self, id: Self::GraphLabel, graph: Self::Graph) -> Option<Self::Graph>;
 
 	/// Insert a quad in the dataset.
-	fn insert(&mut self, quad: Quad<T, T, T, T>);
+	fn insert(
+		&mut self,
+		quad: Quad<Self::Subject, Self::Predicate, Self::Object, Self::GraphLabel>,
+	);
 
 	/// Absorb the given other dataset.
 	///
 	/// Adds all the quads of `other` in the dataset.
-	fn absorb<D: SizedDataset<T>>(&mut self, other: D)
-	where
-		D::Graph: SizedGraph<T>;
+	fn absorb<
+		D: SizedDataset<
+			Subject = Self::Subject,
+			Predicate = Self::Predicate,
+			Object = Self::Object,
+			GraphLabel = Self::GraphLabel,
+		>,
+	>(
+		&mut self,
+		other: D,
+	) where
+		D::Graph: SizedGraph;
 }
